' This program is free software; you can redistribute it and/or
' modify it under the terms of the GNU General Public License
' as published by the Free Software Foundation; either version 2
' of the License, or (at your option) any later version.
'
' This program is distributed in the hope that it will be useful,
' but WITHOUT ANY WARRANTY; without even the implied warranty of
' MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
' GNU General Public License for more details.
'
' You should have received a copy of the GNU General Public License
' along with this program; if not, write to the Free Software
' Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Private Const INDEX_OUT_OF_BOUNDS = 9 '! Error number for accesses outside an array's boundaries.
                                      '! @see http://msdn.microsoft.com/en-us/library/xe43cc8d.aspx

'! A class for more sophisticated array handling than the VBScript built-in
'! arrays allow for. Of course there's System.Collections.ArrayList, but I was
'! aiming for something that a) would not require .NET and b) could easily be
'! used as a queue or stack as well as a "regular" array.
'!
'! @author  Ansgar Wiechers <ansgar.wiechers@planetcobalt.net>
'! @date    2011-01-02
'! @version 1.0
'!
'! @see http://msdn.microsoft.com/en-us/library/system.collections.arraylist.aspx
Class CArray
	Private array_  '! Member variable that contains the actual array.
	Private type_   '! Member variable that contains the type name of the class.

	'! Initialize the class with an empty array and set the type name.
	Private Sub Class_Initialize()
		array_ = Array()
		type_  = TypeName(Me)
	End Sub

	'! Get or set the first element of the CArray. If no first element exists
	'! (i.e. the CArray does not contain any elements), an error is raised.
	'! Getting the first element does not remove the element from the CArray.
	'!
	'! @raise Index out of bounds. (9)
	'!
	'! @see #InsertFirst()
	'! @see #RemoveFirst()
	Public Property Get First()
		Assign First, Me.Item(0)
	End Property

	Public Property Let First(val)
		Me.Item(0) = val
	End Property

	Public Property Set First(obj)
		Set Me.Item(0) = obj
	End Property

	'! Get or set the last element of the CArray. If no last element exists
	'! (i.e. the CArray does not contain any elements), an error is raised.
	'! Getting the last element does not remove the element from the CArray.
	'!
	'! @raise Index out of bounds. (9)
	'!
	'! @see #Push()
	'! @see #Pop()
	'! @see #InsertLast()
	'! @see #RemoveLast()
	Public Property Get Last()
		Assign Last, Me.Item(UBound(array_))
	End Property

	Public Property Let Last(val)
		Me.Item(UBound(array_)) = val
	End Property

	Public Property Set Last(obj)
		Set Me.Item(UBound(array_)) = obj
	End Property

	'! Get or set the element at the given index of the CArray. If the index is
	'! outside the CArray's boundaries, an error is raised. Getting an element
	'! does not remove the element from the CArray.
	'!
	'! @param  index  Index of the element to get/set.
	'!
	'! @raise Index out of bounds. (9)
	'!
	'! @see #InsertBefore()
	'! @see #Remove()
	Public Default Property Get Item(index)
		If index < 0 Or index > UBound(array_) Then Err.Raise _
			INDEX_OUT_OF_BOUNDS, WScript.ScriptName, "Index out of bounds."
		If IsArray(array_(index)) Then
			Set Item = New CArray
			Item.Items = array_(index)
		Else
			Assign Item, array_(index)
		End If
	End Property

	Public Property Let Item(index, val)
		If index < 0 Or index > UBound(array_) Then Err.Raise _
			INDEX_OUT_OF_BOUNDS, WScript.ScriptName, "Index out of bounds."
		array_(index) = val
	End Property

	Public Property Set Item(index, obj)
		If index < 0 Or index > UBound(array_) Then Err.Raise _
			INDEX_OUT_OF_BOUNDS, WScript.ScriptName, "Index out of bounds."
		If TypeName(obj) = type_ Then
			array_(index) = obj.ToArray
		Else
			Set array_(index) = obj
		End If
	End Property

	'! Initialize the CArray with data or get all elements from the CArray. When
	'! (Re-)initializing the CArray with (new) data, any data already present in
	'! the CArray is discarded. Data that is neither an array nor a CArray is
	'! treated as an array with size 1.
	Public Property Get Items()
		Set Items = New CArray
		Items.Items = array_
	End Property

	Public Property Let Items(val)
		If IsArray(val) Then
			array_ = val
		Else
			array_ = Array(val)
		End If
	End Property

	Public Property Set Items(obj)
		If TypeName(obj) = type_ Then
			array_ = obj.ToArray
		Else
			array_ = Array(obj)
		End If
	End Property

	'! Insert the given element as the first element. This is basically an alias
	'! for InsertBefore(0, element) to have coherent method names for using
	'! CArray as a queue.
	'!
	'! @param  element  Element to insert into the CArray.
	'!
	'! @see #InsertLast()
	'! @see #RemoveFirst()
	'! @see #RemoveLast()
	'! @see #InsertBefore()
	Public Sub InsertFirst(element)
		Me.InsertBefore 0, element
	End Sub

	'! Insert the given element as the last element. This is basically an alias
	'! for Add() to have coherent method names for using CArray as a queue.
	'!
	'! @param  element  Element to add to the CArray.
	'!
	'! @see #InsertFirst()
	'! @see #RemoveFirst()
	'! @see #RemoveLast()
	'! @see #Add()
	Public Sub InsertLast(element)
		Me.Add element
	End Sub

	'! Remove the first element from the CArray and return it to the caller. If
	'! the CArray does not contain any elements, an error is raised. This is
	'! basically an alias for Remove(0) to have coherent method names for using
	'! CArray as a queue.
	'!
	'! @return The element that was first in the CArray.
	'!
	'! @raise Index out of bounds. (9)
	'!
	'! @see #InsertFirst()
	'! @see #InsertLast()
	'! @see #RemoveLast()
	'! @see #Remove()
	Public Function RemoveFirst()
		Assign RemoveFirst, Me.Remove(0)
	End Function

	'! Remove the first element from the CArray and return it to the caller. If
	'! the CArray does not contain any elements, an error is raised. This is
	'! basically an alias for Remove(CArray.UpperBound) to have coherent method
	'! names for using CArray as a queue.
	'!
	'! @return The element that was last in the CArray.
	'!
	'! @raise Index out of bounds. (9)
	'!
	'! @see #InsertFirst()
	'! @see #InsertLast()
	'! @see #RemoveFirst()
	'! @see #Remove()
	Public Function RemoveLast()
		Assign RemoveLast, Me.Remove(UBound(array_))
	End Function

	'! Insert the given element as the last element. This is basically an alias
	'! for Add() to have coherent method names for using CArray as a stack.
	'!
	'! @param  element  Element to add to the CArray.
	'!
	'! @see #Pop()
	'! @see #Add()
	Public Sub Push(element)
		Me.Add element
	End Sub

	'! Remove the last element from the CArray and return it to the caller. If
	'! the CArray does not contain any elements, an error is raised. This is
	'! basically an alias for Remove(CArray.UpperBound) to have coherent method
	'! names for using CArray as a stack.
	'!
	'! @return The element that was last in the CArray.
	'!
	'! @raise Index out of bounds. (9)
	'!
	'! @see #Push()
	'! @see #Remove()
	Public Function Pop()
		Assign Pop, Me.Remove(UBound(array_))
	End Function

	'! Insert the given element into the CArray before the given index.
	'!
	'! @param  index    Index before which the element should be inserted.
	'! @param  element  Element to insert into the CArray.
	'!
	'! @raise Index out of bounds. (9)
	Public Sub InsertBefore(index, element)
		Dim i

		ReDim Preserve array_(UBound(array_)+1)
		For i = UBound(array_)-1 To index Step -1
			Assign array_(i+1), array_(i)
		Next

		If TypeName(element) = type_ Then
			Assign array_(index), element.ToArray
		Else
			Assign array_(index), element
		End If
	End Sub

	'! Add the given element at the end of the CArray. Array and CArray elements
	'! are added as nested items. To concatenate an array or CArray with this
	'! CArray, use Append() instead.
	'!
	'! @param  element  The element to add.
	'!
	'! @see #Append()
	Public Sub Add(element)
		ReDim Preserve array_(UBound(array_)+1)
		If TypeName(element) = type_ Then
			Assign array_(UBound(array_)), element.ToArray
		Else
			Assign array_(UBound(array_)), element
		End If
	End Sub

	'! Append each element of the given array or CArray to the end of this
	'! CArray. Arguments that are not an array or CArray are simply added to
	'! the CArray (you could as well use Add() in this case).
	'!
	'! @param  arr  The array or CArray to append.
	'!
	'! @see #Add()
	Public Sub Append(ByVal arr)
		Dim i, start

		If TypeName(arr) = type_ Then arr = arr.ToArray

		If IsArray(arr) Then
			start = UBound(array_) + 1
			ReDim Preserve array_(start + UBound(arr))
			For i = 0 To UBound(element)
				Assign array_(start+i), arr(i)
			Next
		Else
			Me.Add arr
		End If
	End Sub

	'! Remove the element at the given index from the CArray and return it to the
	'! caller. If the given index is outside the boundaries of the CArray, an
	'! error is raised.
	'!
	'! @param  index  The index of the element to remove.
	'! @return The element that was removed from the CArray.
	'!
	'! @raise Index out of bounds. (9)
	Public Function Remove(index)
		Dim i

		If index < 0 Or index > UBound(array_) Then Err.Raise _
			INDEX_OUT_OF_BOUNDS, WScript.ScriptName, "Index out of bounds."

		If IsArray(array_(index)) Then
			Set Remove = New CArray
			Remove.Items = array_(index)
		Else
			Assign Remove, array_(index)
		End If

		For i = index+1 To UBound(array_)
			array_(i-1) = array_(i)
		Next
		ReDim Preserve array_(UBound(array_)-1)   ' truncate array_
	End Function

	'! Return a slice (sub-array) from this CArray. The interval defined by the
	'! indexes is left-closed right-open (i.e. the element referred to by the
	'! last index is not included with the slice).
	'!
	'! If both index values are outside the CArray's boundaries, or if the last
	'! index is greater than or equal to the first index, an empty CArray is
	'! returned. Negative indexes are allowed and refer to elements counted from
	'! the upper boundary of the array, starting with -1. Parameters can be Null
	'! and are then interpreted as the CArray's lower or upper boundary
	'! respectively.
	'!
	'! @param  first  Index of the beginning of the slice (included).
	'! @param  last   Index of the end of the slice (excluded).
	'! @return A slice from the CArray. The slice is also a CArray.
	Public Function GetSlice(ByVal first, ByVal last)
		Dim a, i

		a = Array()

		' convert Null parameters into the respective indexes
		If IsNull(first) Then first = 0
		If IsNull(last) Then last = Me.Count

		' convert negative indexes into "normal" indexes
		If first < 0 Then first = first + Me.Count
		If last < 0 Then last = last + Me.Count

		' interval is right-open, so last index is not included
		last = last - 1

		' slice cannot contain values from arr if any of these are true:
		' - first > last
		' - both first and last < first index of arr
		' - both first and last > last index of arr
		If first <= last And last >= 0 And first <= UBound(array_) Then
			If first < 0 Then first = 0
			If last > UBound(array_) Then last = UBound(array_)

			Redim a(last-first)

			For i = first To last
				assign a(i-first), array_(i)
			Next
		End If

		Set GetSlice = New CArray
		GetSlice.Items = a
	End Function

	'! Remove a slice (sub-array) from this CArray and return it to the caller.
	'! The interval defined by the indexes is left-closed right-open (i.e. the
	'! element referred to by the last index is not included with the slice).
	'!
	'! If both index values are outside the CArray's boundaries, or if the last
	'! index is greater than or equal to the first index, nothing is removed and
	'! an empty CArray is returned. Negative indexes are allowed and refer to
	'! elements counted from the upper boundary of the array, starting with -1.
	'! Parameters can be Null and are then interpreted as the CArray's lower or
	'! upper boundary respectively.
	'!
	'! @param  first  Index of the beginning of the slice (included).
	'! @param  last   Index of the end of the slice (excluded).
	'! @return The slice removed from the CArray. This slice is also a CArray.
	Public Function RemoveSlice(ByVal first, ByVal last)
		Dim a, i, offset, ca

		a = Array()

		' convert Null parameters into the respective indexes
		If IsNull(first) Then first = 0
		If IsNull(last) Then last = Me.Count

		' convert negative indexes into "normal" indexes
		If first < 0 Then first = first + Me.Count
		If last < 0 Then last = last + Me.Count

		' interval is right-open, so last index is not included
		last = last - 1

		' slice cannot contain values from arr if any of these are true:
		' - first > last
		' - both first and last < first index of arr
		' - both first and last > last index of arr
		If first <= last And last >= 0 And first <= UBound(array_) Then
			If first < 0 Then first = 0
			If last > UBound(array_) Then last = UBound(array_)

			' copy slice elements to array a
			Redim a(last-first)
			For i = first To last
				assign a(i-first), array_(i)
			Next

			' move remaining elements to front
			offset = last - first + 1
			For i = last+1 To UBound(array_)
				assign array_(i-offset), array_(i)
			Next

			' truncate array_
			ReDim Preserve array_(UBound(array_)-offset)
		End If

		Set RemoveSlice = New CArray
		RemoveSlice.Items = a
	End Function

	'! Return the elements of the CArray as a plain VBScript array.
	'!
	'! @return Array with the elements of the CArray.
	Public Function ToArray()
		ToArray = array_
	End Function

	'! Return the number of elements in this CArray.
	'!
	'! @return The number of elements in the CArray.
	Public Function Count()
		Count = UBound(array_) + 1
	End Function

	'! Return the index of this CArray's upper boundary.
	'!
	'! @return The index of the CArray's upper boundary.
	Public Function UpperBound()
		UpperBound = UBound(array_)
	End Function

	'! Remove all elements from the CArray.
	Public Sub Clear()
		array_ = Array()
	End Sub

	'! Transform a multi-dimensional CArray into a one-dimensional CArray by
	'! expanding nested arrays.
	Public Sub Flatten()
		Expand array_
	End Sub

	'! Worker function to recursively expand nested arrays.
	'!
	'! @param  arr  The array whose nested arrays should be expanded.
	Private Sub Expand(arr)
		Dim i, j, a

		' Can't use a for loop here, because UBound(arr) will probably grow in the
		' process.
		i = 0
		Do While i <= UBound(arr)
			If IsArray(arr(i)) Then
				' flatten nested array first
				Expand arr(i)

				' resize "parent" array and move elements after index i to the end
				offset = UBound(arr(i))
				ReDim Preserve arr(UBound(arr) + offset)
				For j = UBound(arr) To UBound(arr)-offset+1 Step -1
					Assign arr(j), arr(j-offset)
				Next

				' expand nested array at index i
				a = arr(i)
				For j = UBound(a) To 0 Step -1
					Assign arr(i+j), a(j)
				Next
			End If

			i = i + 1
		Loop
	End Sub

	'! Return a CArray with all indexes on which the given element occurs in the
	'! CArray. The function does not recurse into nested arrays.
	'!
	'! @param  element  The element for which to check.
	'! @return CArray with all indexes for the given element.
	Public Function AllIndexOf(element)
		Dim i, compareObjects

		compareObjects = IsObject(element)
		Set AllIndexOf = New CArray

		For i = 0 To UBound(array_)
			If compareObjects Then
				If IsObject(array_(i)) Then
					If element Is array_(i) Then AllIndexOf.Add i
				End If
			Else
				If Not IsObject(array_(i)) And Not IsArray(array_(i)) Then
					If element = array_(i) Then AllIndexOf.Add i
				End If
			End If
		Next
	End Function

	'! Return the index of the first occurrance of the given element in the
	'! CArray. If no occurrance is found, the return value is Null. The function
	'! does not recurse into nested arrays.
	'!
	'! @param  element  The element to check for.
	'! @return The index of the first occurrance of element or Null.
	Public Function IndexOf(element)
		Dim i, compareObjects

		compareObjects = IsObject(element)

		For i = 0 To UBound(array_)
			IndexOf = i
			If compareObjects Then
				If IsObject(array_(i)) Then
					If element Is array_(i) Then Exit Function
				End If
			Else
				If Not IsObject(array_(i)) And Not IsArray(array_(i)) Then
					If element = array_(i) Then Exit Function
				End If
			End If
		Next

		IndexOf = Null
	End Function

	'! Return the index of the last occurrance of the given element in the
	'! CArray. If no occurrance is found, the return value is Null. The function
	'! does not recurse into nested arrays.
	'!
	'! @param  element  The element to check for.
	'! @return The index of the last occurrance of element or Null.
	Public Function LastIndexOf(element)
		Dim i, compareObjects

		compareObjects = IsObject(element)
		IndexOf = Null

		For i = 0 To UBound(array_)
			If compareObjects Then
				If IsObject(array_(i)) Then
					If element Is array_(i) Then IndexOf = i
				End If
			Else
				If Not IsObject(array_(i)) And Not IsArray(array_(i)) Then
					If element = array_(i) Then IndexOf = i
				End If
			End If
		Next
	End Function

	'! Indicate whether the given element does or doesn't exist in this CArray.
	'! The function does not recurse into nested arrays.
	'!
	'! @param  element  The element to check for.
	'! @return True if the element is found in the CArray, otherwise False.
	Public Function Exists(element)
		Dim i, compareObjects

		compareObjects = IsObject(element)
		Exists = False

		For i = 0 To UBound(array_)
			If compareObjects Then
				If IsObject(array_(i)) Then
					If element Is array_(i) Then
						Exists = True
						Exit Function
					End If
				End If
			Else
				If Not IsObject(array_(i)) And Not IsArray(array_(i)) Then
					If element = array_(i) Then
						Exists = True
						Exit Function
					End If
				End If
			End If
		Next
	End Function

	'! Revert the order of the elements in the CArray.
	Public Sub Reverse()
		Dim a, i

		a = array_
		For i = 0 To UBound(array_)
			Assign array_(i), a(UBound(array_)-i)
		Next
	End Sub

	'! Sort the CArray in ascending order. If the CArray contains any complex
	'! data types (like nested arrays or objects) an error will be raised. Arrays
	'! will raise error 10 (type mismatch), while objects will raise error 438
	'! (property or method not supported).
	'!
	'! @raise Type mismatch. (10)
	'! @raise Object doesn't support this property or method. (438)
	'!
	'! @see http://msdn.microsoft.com/en-us/library/xe43cc8d.aspx
	Public Sub Sort()
		QuickSort array_, 0, UBound(array_)
	End Sub

	'! Sort a given array in ascending order using the quicksort algorithm. If
	'! array contains any complex data types (like nested arrays or objects) an
	'! error will be raised. Arrays will raise error 10 (type mismatch), while
	'! objects will raise error 438 (property or method not supported).
	'!
	'! @param  arr    The array to sort.
	'! @param  left   Left (lower) boundary of the array slice the current
	'!                recursion step will operate on.
	'! @param  right  Right (upper) boundary of the array slice the current
	'!                recursion step will operate on.
	'!
	'! @raise Type mismatch. (10)
	'! @raise Object doesn't support this property or method. (438)
	'!
	'! @see http://en.wikipedia.org/wiki/Quicksort
	'! @see http://msdn.microsoft.com/en-us/library/xe43cc8d.aspx
	Private Sub QuickSort(arr, left, right)
		Dim pivot, leftIndex, rightIndex, buffer

		leftIndex = left
		rightIndex = right

		If right - left > 0 Then
			pivot = Int((left + right) / 2)

			While leftIndex <= pivot And rightIndex >= pivot
				On Error Resume Next
				While arr(leftIndex) < arr(pivot) And leftIndex <= pivot
					leftIndex = leftIndex + 1
				Wend
				While arr(rightIndex) > arr(pivot) And rightIndex >= pivot
					rightIndex = rightIndex - 1
				Wend

				Assign buffer, arr(leftIndex)
				Assign arr(leftIndex), arr(rightIndex)
				Assign arr(rightIndex), buffer

				leftIndex = leftIndex + 1
				rightIndex = rightIndex - 1
				If leftIndex - 1 = pivot Then
					rightIndex = rightIndex + 1
					pivot = rightIndex
				ElseIf rightIndex + 1 = pivot Then
					leftIndex = leftIndex - 1
					pivot = leftIndex
				End If
			Wend

			QuickSort arr, left, pivot-1
			QuickSort arr, pivot+1, right
		End If
	End Sub

	'! Assign a value or object reference to a variable. A helper function, to
	'! handle primitive data types and objects alike.
	'!
	'! @param  var  Reference to a variable, to which val will be assigned.
	'! @param  val  value or reference to assign to var.
	Private Sub Assign(ByRef var, val)
		If IsObject(val) Then
			Set var = val
		Else
			var = val
		End If
	End Sub
End Class
